import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;

public class Sort {
  private static long currentTimerStart = System.nanoTime();

  /**
   * You are to implement this method. The method should invoke one or
   * more threads to read and sort the data from the collection of Files.
   * The method should return a sorted list of all of the String data
   * contained in the files.
   *
   * @param files array of files to read strings from.
   * @return sorted array
   * @throws IOException
   */
  public static String[] threadedSort(File[] files) throws IOException {
    final int nthreads = Integer.parseInt(System.getProperty("SORT_NTHREADS"));

    String[] allData = new String[0];

    long time;
      time = System.nanoTime();

    for (File file : files) {
      String[] data = getData(file);
      allData = concatenateArrays(allData, data);
    }

    System.out.println("Retrieving data: " + updateTimer());

    return threadedSortInternal(allData, nthreads);
  }


  /**
   * Given an array of files, this method will return a sorted
   * list of the String data contained in each of the files.
   *
   * @param files the files to be read
   * @return the sorted data
   * @throws IOException thrown if any errors occur reading the file
   */
  public static String[] sort(File[] files) throws IOException {

    String[] sortedData = new String[0];

    for (File file : files) {
      String[] data = getData(file);
      data = MergeSort.mergeSort(data);
      sortedData = MergeSort.merge(sortedData, data);
    }

    return sortedData;

  }


  /**
   * This method will read in the string data from the specified
   * file and return the data as an array of String objects.
   *
   * @param file the file containing the String data
   * @return String array containing the String data
   * @throws IOException thrown if any errors occur reading the file
   */
  private static String[] getData(File file) throws IOException {
    ArrayList<String> data = new ArrayList<String>();
    //changes made on the file read and buffering
    FileReader fr = new FileReader(file);
    BufferedReader in = new BufferedReader(fr);

    // Read the data from the file until the end of file is reached
    while (in.readLine() != null) {
      String line = in.readLine();
      if (line == null) {
        // the end of file was reached
        break;
      } else {
        data.add(line);
      }
    }

    //Close the input stream and return the data
    in.close();
    return data.toArray(new String[0]);
  }

  private static class SortThreadRunner implements Runnable {
    private final String[] lines;

    SortThreadRunner(String[] lines) {
      this.lines = lines;
    }

    @Override
    public void run() {
      String[] sortedLines = MergeSort.mergeSort(lines);
      System.arraycopy(sortedLines, 0, lines, 0, lines.length);
    }
  }

  private static String[] threadedSortInternal(String[] lines, int nthreads) {
    if (lines.length < nthreads) {
      return MergeSort.mergeSort(lines);
    }

    final int maxThreadLinesLength = (lines.length + nthreads - 1) / nthreads;
    String[][] threadLines = new String[nthreads][];
    for (int i = 0; i < nthreads; i++) {
      // starting and finishing position for i-th thread to sort from.
      final int thisThreadStart = i * maxThreadLinesLength;
      int thisThreadFinish = thisThreadStart + maxThreadLinesLength;
      if (thisThreadFinish > lines.length) {
        thisThreadFinish = lines.length;
      }

      threadLines[i] = Arrays.copyOfRange(lines, thisThreadStart, thisThreadFinish);
    }

    System.out.println("Broadcasting array among threads: " + updateTimer());

    // Initializing and starting threads.
    final Thread[] threads = new Thread[nthreads];
    for (int i = 0; i < nthreads; i++) {
      threads[i] = new Thread(new SortThreadRunner(threadLines[i]));
      threads[i].start();
    }

    System.out.println("started threads: " + updateTimer());

    // Wait until all sorting threads are finished.
    for (int i = 0; i < nthreads; i++) {
      try {
        threads[i].join();
      } catch (InterruptedException e) {
        throw new RuntimeException(e);
      }
    }

    System.out.println("joined threads: " + updateTimer());

    // merging stage: main thread merges all the arrays.
    String[] sortedData = new String[0];
    for (int i = 0; i < nthreads; i++) {
      String[] oldSortedData = sortedData;
      sortedData = MergeSort.merge(oldSortedData, threadLines[i]);
    }

    System.out.println("merged lines: " + updateTimer());

    return sortedData;
  }

  /**
   * Concatenates two array of String type.
   * @param lhs left hand side array.
   * @param rhs right hand side array.
   * @return concatenation of lhs and rhs with appropriate size.
   */
  private static String[] concatenateArrays(String[] lhs, String[] rhs) {
    String[] result = Arrays.copyOf(lhs, lhs.length + rhs.length);
    System.arraycopy(rhs, 0, result, lhs.length, rhs.length);
    return result;
  }

  /**
   *
   * @return Time passed since the last call of updateTimer (in seconds).
   */
  private static double updateTimer() {
    double result = (System.nanoTime() - currentTimerStart) / 1e9;
    currentTimerStart = System.nanoTime();
    return result;
  }
}
